#!/bin/bash

#This script watches for new files written in a local raw dir and
#moves them to HDFS.

# LANDER_RAWDIR         : where dagtalker writes data (HDFS: local, NFS: NFS)
# LANDER_QUEUE_RAW_HDFS : where to store stuff in HDFS
# LANDER_QUEUE_TYPE     : exit if this is not set to 'hdfs'

LANDER_CONFIG=~plander/lander.conf
#read config
if [ -e $LANDER_CONFIG ]; then
    . $LANDER_CONFIG
else
    echo "ERROR: Config file $CONFIG does not exist" >&2
    exit 1
fi

[ "$LANDER_QUEUE_TYPE" == "hdfs" ] || {
    echo "LANDER_QUEUE_TYPE=$LANDER_QUEUE_TYPE, nothing to do"
    exit 0 #not hdfs, nothing to do
}

HDFSMOVERLOG=~plander/hdfsmover.log
HDFSMAXTRIES=5
LANDER_HDFSMOVER_MAXPARALLEL=5

exec >> $HDFSMOVERLOG 2>&1

#Allow starting this script from cron job, ensure only one
#+instance is running
HDFSMOVER_PID=$HOME/hdfsmover.pid
if [ -e $HDFSMOVER_PID ]; then
    read pid rest < $HDFSMOVER_PID
    if [ -n "$pid" ] && kill -0 $pid; then
	info "hdfsmover is already running as $pid"
	exit 0
    fi
fi
echo $$ > $HDFSMOVER_PID

eval HDFS="~plander/hadoop/bin/hadoop\ dfs"

LOGERROR=1
LOGWARN=2
LOGINFO=3

LOGLEVEL=$LOGINFO
ADMIN=yuri@isi.edu

log() {
    log="$(date --rfc-3339=s)\t$@"
    echo -e "$log"
}

logmail() {
    log "$@"
    mail -s "hdfsmover error" $ADMIN <<EOF

hdfsmover error on $(hostname): 
    $(echo -e "$log")

EOF
}

info() {
    [ $LOGLEVEL -ge $LOGINFO ] && log "INFO\t$@"
}
warn() {
    [ $LOGLEVEL -ge $LOGWARN ] && log "WARN\t$@"
}
error() {
    [ $LOGLEVEL -ge $LOGERROR ] && logmail "ERROR\t$@"
}

hdfs_rename_file() {
    local _s=$1 _d=$2
    local _try

    for _try in {1..$HDFSMAXTRIES}; do
	if $HDFS -mv $_s $_d; then
	    info "moved $_s to $_d OK"
	    return 0
	fi
	$HDFS -rm $_d >& /dev/null
    done
    error "Failed to rename $_s to $_d after $HDFSMAXTRIES"
    return 1
}

#deletes source from fs only if successful, otherwise removes destination
hdfs_copyin_file() {
    local _s=$1 _d=$2
    local _try
    for _try in {1..$HDFSMAXTRIES}; do
	if $HDFS -put $_s $_d; then
	    info "copied $_s to $_d OK"
	    rm -f $_s
	    return 0
	fi
	#remove dst and retry
	$HDFS -rm $_d 2>/dev/null
    done
    error "Failed to copyin $_s to $_d after $_try attempts"
    return 1
}

#forks at most $LANDER_HDFSMOVER_MAXPARALLEL hdfs -put's
#if so many are already running, sleeps, polling.  Uses job control.
hdfs_moverjob() {
    local _s=$1 _d=$2
    while [ $(jobs -pr | wc -l) == "$LANDER_HDFSMOVER_MAXPARALLEL" ]; do
	sleep 1
    done
    { 
	hdfs_copyin_file "$_s" "$_d.copyin" && 
	hdfs_rename_file "$_d.copyin $_d" || 
	sleep 5s
    } &
}
    
while true; do
    #check if there are any stale .copyin files and rename them
    lsout=$($HDFS -ls $LANDER_QUEUE_RAW_HDFS*/*.copyin* 2>/dev/null)
    if [ $? == 0 ]; then
	if [ -n "$lsout" ]; then
	    #there are stale files to move
	    stalefiles=$(echo "$lsout" | grep .copyin | sed 's/.* //')
	    info "There are " $(echo "$stalefiles" | wc -l) " stale files to rename"
	    for f in $stalefiles; do 
		hdfs_rename_file $f ${f%.copyin*} || continue 2
	    done
	fi
    else
	if ! grep -iq 'no such file or directory' <<<$lsout; then
	    error "Can't get a listing of HDFS directory"
	    sleep 5m
	    continue
	fi
    fi
    
    
    #start moving new files
    files=$(find $LANDER_RAWDIR/ -regextype posix-extended \
	-regex '.*\/[0-9]{8}-[0-9]{6}-[0-9]+$' -mmin +5 -print 2>/dev/null)
    [ -z "$files" ] && { sleep 30s; continue; }

    # sort files by name (but ignore the directories)
    fnkey=$(echo "/x$files" 2>/dev/null | head -1 | tr '/' ' ' | wc -w) #should be num of '/' + 1
    sortedfiles=$(echo "$files" | sort -t '/' -k $fnkey)

    info "Have " $(echo "$sortedfiles" | wc -l) " to move"

    for FULLFILE in $sortedfiles; do
	lander=$(basename $(dirname $FULLFILE))
	file=$(basename $FULLFILE)
	hFULLFILE=$LANDER_QUEUE_RAW_HDFS/$lander/$file
	
	hdfs_moverjob $FULLFILE $hFULLFILE #this parallelizes hdfs -put

    done #for FULLFILE in $sortedfiles
done #while true

error "Reach the end of while...true"
sleep 1m
exec $0
