#!/bin/bash  
#PBS -N scrubber_worker.sh 
#PBS -l walltime=2:00:00,mem=4gb
#PBS -k n
#PBS -M plander@hpc.usc.edu
#PBS -o hpc-master:/auto/rcf-26/yuri/scrubber_worker_job.out
#PBS -e hpc-master:/auto/rcf-26/yuri/scrubber_worker_job.err
#xxx -o hpc-master:/home/rcf-15/plander/logs/scrubber_worker_job.out
#xxx -e hpc-master:/home/rcf-15/plander/logs/scrubber_worker_job.err
#xxx -vx above
# $Id: scrubber_worker.sh 15973 2013-02-14 21:42:39Z yuri $

THISSCRIPT=$0

LANDER_CONFIG=~plander/lander.conf 
#read config
if [ -e $LANDER_CONFIG ]; then
    . $LANDER_CONFIG
else
    echo "ERROR: Config file $LANDER_CONFIG does not exist" >&2
    exit 1
fi

# throw away all output, log only interesting things
exec >> $LANDER_SCRUBBER_WORKER_LOGFILE  2>&1

WORKERID="$(hostname -s)-$$"

worker_args=${WORKER_ARGS:-$@}

jobspercore=$LANDER_SCRUBBER_JOBSPERCORE
if [ -n "$jobspercore" ] && [ -z "$LANDER_WORKER_FORK" ]; then
    CPUINFO=/proc/cpuinfo
    if [ -e $CPUINFO ]; then
	ncores=$(grep ^process $CPUINFO | wc -l)
	njobs=$((ncores*jobspercore/100))
	njobs=$(( (njobs < 1)?1:njobs ))
	export LANDER_WORKER_FORK="yes"
	forks=0
	for (( jobs=0; jobs<njobs; ++jobs )); do
	    $THISSCRIPT $worker_args &
	    let '++forks'
	done
	echo "*** forked $forks times" >&2
	wait
	echo "all forks completed" >&2
	exit 0
    else
	echo "$CPUINFO file doesn't exist, ignoring LANDER_SCRUBBER_JOBSPERCORE config option" >&2;
    fi
fi

set -- $worker_args
LANDER_RAWDIR_FS=$LANDER_RAWDIR_FS/$1     #should work with empty args xxx
LANDER_RAWDIR_HDFS=$LANDER_RAWDIR_HDFS/$1

LS=/bin/ls
RM=/bin/rm
MV=/bin/mv
READLINK="readlink -f"

#preemptive exit via signalling
must_exit=""
trap "must_exit=y" SIGHUP SIGQUIT SIGABRT SIGSTOP SIGTERM

modtime_file() {
    local _file=$1
    /usr/bin/stat --format %Y $_file 2>/dev/null
}

#use modtime vars to be sure we always have the latest user lists
modtime_q=-1 #quota
users_r=
users_h=
users_l=
users_f=

for hook in $LANDER_RAW_HOOKS; do
    eval users_hook_${hook}=
done

update_data_users() {
    local _mq=$(modtime_file $LANDER_QUOTAFILE)
    if [ $_mq -gt $modtime_q ]; then
	modtime_q=$_mq
	#silently rebuild all lists of active users
	users_r=$(getent group $LANDER_GROUP_RAW)
	users_h=$(getent group $LANDER_GROUP_HOST)
	users_l=$(getent group $LANDER_GROUP_FLOW)
	users_f=$(getent group $LANDER_GROUP_FULL)

	lander_active_users raw
	lander_active_users host
	lander_active_users flow
	lander_active_users full
	for hook in $LANDER_RAW_HOOKS; do
	    eval local _group=\$LANDER_GROUP_$(echo $hook | tr '[:lower:]' '[:upper:]')
	    eval users_hook_${hook}=\$\(getent group $_group\)
	    lander_active_users $hook
	done
    else
	# see if group membership has changed
	if [ "$(getent group $LANDER_GROUP_RAW)"  != "$users_r" ]; then
	    echo "INF group membership changed  (raw), reloading" 
	    users_r=$(getent group $LANDER_GROUP_RAW)
	    lander_active_users raw
        fi
	if [ "$(getent group $LANDER_GROUP_HOST)" != "$users_h" ]; then
	    echo "INF group membership changed (host), reloading" 
	    users_h=$(getent group $LANDER_GROUP_HOST)
	    lander_active_users host
        fi
	if [ "$(getent group $LANDER_GROUP_FLOW)" != "$users_l" ]; then
	    echo "INF group membership changed (flow), reloading" 
	    users_l=$(getent group $LANDER_GROUP_FLOW)
	    lander_active_users flow
        fi
	if [ "$(getent group $LANDER_GROUP_FULL)" != "$users_f" ]; then
	    echo "INF group membership changed (full), reloading" 
	    users_f=$(getent group $LANDER_GROUP_FULL)
	    lander_active_users full
        fi
	for hook in $LANDER_RAW_HOOKS; do
	    eval local _group=\$LANDER_GROUP_$(echo $hook | tr '[:lower:]' '[:upper:]')
	    eval users_hook=\$users_hook_${hook}
	    if [ "$(getent group $_group)" != "$users_hook" ]; then
		echo "INF group membership changed ($hook:$_group), reloading"
		eval users_hook_${hook}=\$\(getent group $_group\)
		lander_active_users $hook
	    fi
	done
    fi
}
	
process_queue() {
    local _type=$1 _inpfile=$2 _users=$3 _userfile= _u=
    if process_$_type "$_inpfile"; then
	local _pob=$(basename $process_out)
	#distribute files
	_userfile="$LANDER_TMPDIR_U/$DIR/$_pob-$WORKERID"
	chmod 660 "$process_out"
	ln "$process_out" "$_userfile" 2>/dev/null || cp -p "$process_out" "$_userfile" || \
	    { echo "ERR copying of $_type output failed"; return 1; }
	chmod 660 "$_userfile" #just in case
	for _u in $_users;  do
	    lander_user_subscribed $_u $_type $DIR || continue
	    ln "$_userfile" "$LANDER_TOPDIR_U/${_u}/incoming_${_type}/$DIR/$_pob" || \
                echo "User $user did not get $_type data"
	done
	$RM -f "$_userfile"
	cleanup_local+="$process_out "
	return 0
    fi
    echo "ERR process_$type failed on ($DIR) $_inpfile"
    cleanup_local+="$process_out "
    return 1 
}

process_raw() {
    local _inpfile=$1
    process_out="${_inpfile}-r"
    ln "$_inpfile" "$process_out"
    chgrp $LANDER_GROUP_RAW $process_out
    return 0
}

process_flow() {
    # FLOW DATA: host-anonymized with a consistent key processed by argus and
    #+tweaked timestamps
    local _inpfile=$1 _argusout=$1-argus _ka_id=${FLOWKEY##*-}
    cleanup_local+="$_argusout "
    process_out="${_inpfile}-h${_ka_id}.flow"
    if $LANDER_SCRUBBER -l /dev/null $LANDER_SCRUBBER_OPT_HOST -s $LANDER_FLOWKEY $_inpfile |\
        $LANDER_TRACECONVERT erf:- pcap:- |\
        $LANDER_ARGUSBIN -F $LANDER_ARGUSDIR/conf/argus.conf -r- -w $_argusout \
	&& $LANDER_RABIN -F $LANDER_ARGUSDIR/conf/rarc -u -r $_argusout |\
           $LANDER_FLOWDIR/readableTime.py |\
           sort -T /tmp -s -k1,1 >$process_out; then
	#set permissions
	chgrp $LANDER_GROUP_FLOW $process_out
	return 0
    fi
    return 1
}

process_host() {
    # HOST ANON: needs to be done even if there are no host-only data users
    #+get key ids: softlinks may be changed while the script
    #+is still going, so we've got to use the real keys
    local _inpfile=$1 _real_kh= _anon_hopt= _kh_id=
    _real_kh=$($READLINK $LANDER_KEYHOST_FILE)
    _anon_hopt="-s $_real_kh"
    # separate key number
    _kh_id=$(basename $_real_kh)
    _kh_id=${_kh_id##*-}
    process_out="$_inpfile-h${_kh_id}"
    if $LANDER_SCRUBBER -l $LANDER_SCRUBBER_LOGFILE $LANDER_SCRUBBER_OPT_HOST \
        $_anon_hopt "$_inpfile" "$process_out"; then
        # Scrubbing worked
	chgrp $LANDER_GROUP_HOST "$process_out"
	return 0
    fi
    return 1
}

process_full() {
    # FULL ANON: this is conditional (if there are users) and dependend on the success
    #+of the previous step as it takes $outH as input get key ids: softlinks may 
    #+be changed while the script is still going, so we've got to use the real keys.
    local _inpfile=$1 _real_kf= _anon_fopt= _kf_id=
    _real_kf=`$READLINK $LANDER_KEYFULL_FILE`
    _anon_fopt="-s $_real_kf"
    # separate key number
    _kf_id=${_real_kf##*-}
    process_out="${_inpfile}f${_kf_id}"
    if $LANDER_SCRUBBER -l $LANDER_SCRUBBER_LOGFILE $LANDER_SCRUBBER_OPT_FULL \
	    $_anon_fopt "$_inpfile" "$process_out"; then
	# Scrubbing worked
	chgrp $LANDER_GROUP_FULL "$process_out"
	return 0
    fi
    return 1
}

    
#
#MAIN:
#
set -o noclobber #just incase
umask 0077

localtmp=$(mktemp -d "$LANDER_LOCALDIR/$WORKERID.XXXXX") ||  { 
    echo "ERR: can't mktemp $LANDER_LOCALDIR/$WORKERID.XXXXX" >&2
    exit 1
}
chmod 0700 $localtmp
cd $localtmp || { 
    echo "ERR: can't cd to $localtmp" >&2
    exit 1
}

now=`date +%s`
# We quit after $LANDER_SCRUBBER_RUN_SEC
let "TIMEDONE = now + LANDER_SCRUBBER_RUN_SEC"

ret=0
while true; do
    now=`date +%s`
    [ $now -lt $TIMEDONE ] || break
    [ -z "$must_exit" ] || { ret=1; break; }
    #check if the load is too high
    lander_set_LOAD
    if [ -n "$LOAD" ]; then
	    echo "$now Current load: $LOAD"
	    diff=`echo $LANDER_SCRUBBER_MAXLOAD - $LOAD | bc -l`
	    if [ -z "${diff/-*/}" ]; then
		echo "Warning: load limit ($LANDER_SCRUBBER_MAXLOAD) is exceeded, sleeping"
	        #randomize sleeping time a little
		sleep $((40 + RANDOM % 60))
		continue
	    fi
    fi
    #got here, means load is OK
    if lander_queue_reserve_top $localtmp; then
	localraw=$reserved_local
	#WORK on this file
	#isolate basename() and basename(dirname())
	FILE=${localraw##*/}
	DIRNAME=${localraw%/*}
	DIR=${DIRNAME##*/}

	update_data_users

	eval users_rawdata=\$active_raw_$DIR
	eval users_hostdata=\$active_host_$DIR
	eval users_flowdata=\$active_flow_$DIR
	eval users_fulldata=\$active_full_$DIR
	
	users_extra=
	for hook in $LANDER_RAW_HOOKS; do
	    eval hook_users=\$active_${hook}_$DIR
	    eval users_${hook}data=\$hook_users
	    users_extra+="$hook_users "
	done
	#echo "Users for $DIR: ($users_rawdata/$users_hostdata/$users_flowdata/$users_fulldata)"
	[ -z "$users_rawdata$users_hostdata$users_flowdata$users_fulldata$users_extra" ] && { 
	    echo "INF: No users for $localraw"
	    lander_queue_release $localraw
	    lander_wipe_source $DIR
	    continue
	}

	cleanup_local="$localraw "
	process_failed=
	[ -n "$users_rawdata" ] && { 
	    process_queue raw  "$localraw" "$users_rawdata" || process_failed=yes
	}
	#run raw hooks
	for hook in $LANDER_RAW_HOOKS; do
	    eval hook_users=\$users_${hook}data
	    [ -n "$hook_users" ] && {
		eval f="\$(type -t process_$hook)"
		if [ "$f" == "function" ]; then
		    process_queue $hook "$localraw" "$hook_users" || process_failed=yes
		else
		    echo "ERROR: no process_$hook function defined, queue is not served" >&2
		fi
	    }
	done
	[ -n "$users_flowdata" ] && {
	    process_queue flow "$localraw" "$users_flowdata" || process_failed=yes
	}
	[ -n "$users_hostdata$users_fulldata" ] && { 
	    if process_queue host "$localraw" "$users_hostdata"; then
		[ -n "$users_fulldata" ] && { 
		    process_queue full "$process_out" "$users_fulldata" || process_failed=yes
		}
	    else
		process_failed=yes
	    fi
	}
	[ -z "$process_failed" ] && lander_queue_release $localraw
	$RM -f $cleanup_local #$localraw may have been removed by release
    else
        #no more files?
	lander_queue_size || {
	    #no more files?
	    sleep 5s
	    lander_queue_size || break #still no more files
	}
    fi
done #while true

$RM -rf $localtmp
exit $ret
