#!/bin/bash

# $Id: user_dispatcher.sh 15104 2012-11-08 20:05:17Z yuri $

#  This script is meant to be the "front end" for the user running LANDER
#+ jobs.  Typically it is run from user's crontab(5).

#  Usage:
#+	user_dispatcher.sh /home/samfs-00/user/myscript.sh [host|full|raw]
#+ note that the user must specify a full path to myscript and it should be
#+ under /home/samfs-00.
#+ The last argument signifies if the user is getting host anonymized data
#+ or fully anonymized data.  If the user is getting both, two dispatchers
#+ can be scheduled.

LANDER_CONFIG=~plander/lander.conf
#read config
if [ -e $LANDER_CONFIG ]; then
    . $LANDER_CONFIG
else
    echo "ERROR: Config file $CONFIG does not exist" >&2
    exit 1
fi

#arguments
max_jobs=$LANDER_USERJOB_MAXJOBS
max_wait=0
verbose=
while getopts "hi:n:uw:j:r:" opt; do
    case $opt in
        ( h ) usage; exit 0; ;;
        ( i ) INDIR=$OPTARG; ;;
        ( j ) jobspercore=$OPTARG; ;;
        ( n ) max_jobs=$OPTARG; ;;
        ( p ) INPROCESSDIR=$OPTARG; ;;
        ( r ) FILE_REGEX=$OPTARG; ;;
        ( s ) SKIPOOO="yes"; ;;
        ( u ) UNORDERED="yes"; ;;
	( v ) verbose="yes"; ;;
        ( w ) max_wait=$OPTARG; ;;
        ( h|* ) usage; exit 0; ;;
    esac
done
shift $((OPTIND-1))

THISSCRIPT=$0
USERSCRIPT=$1				# arg1: user script (executable)
ANONTYPE=${2:-host}			# arg2: anonymization type := (raw|full|host|flow)

file_regex=${FILE_REGEX:-'[0-9]{8}-[0-9]{6}-[0-9]+[^/]*$'}

usage () {
    cat >&2 <<EOF
    USAGE: $THISSCRIPT <OPTIONS> <user_script> [full|flow|host|raw]"
	where:
               <OPTIONS>        can be:
                 -i <indir>     Incoming directory (if different from incoming_(host|full|raw)
                 -j <jobs>      Jobs per core times 100; i.e. 150 will run 1.5*ncores
                 -n N           Run no more than N concurrently (defaults to 
                                lander.conf:USERJOB_MAXJOBS)
                 -p <tmpdir>    Temp dir (should be on the same FS as <indir>
                 -r <regex>     Regular expression for incoming files
		 -s             Skip out of order files (only useful with -n 1)
                 -u             Unordered: no sequence numbers
                 -v             verbose: do additional sanity checks and log the results
                 -w X           Require that the modtime at the head of the queue be 
                                at least X seconds in the past (defaults to 0)

	       <user_script>    must have full path starting $TOPDIR...
	  [full|flow|host|raw]  is the type of anonymization used on incoming data
	                        (defaults to host)
EOF
}

if [ -z "$LOGNAME" ]; then
	echo "LOGNAME isn't set" >&2
	exit 1
fi

if [ -z "$USERSCRIPT" ]; then
    echo "userscript arguement is missing" >&2
    usage
    exit 1
fi

if [ ! -x "$USERSCRIPT" ]; then
    echo "userscript $USERSCRIPT is not executable" >&2
    exit 1
fi

#check that the script is on the NFS
if [ -n "$LANDER_USERJOB_NFSCHECK" ]; then
    t=$(df -tnfs -PT "$USERSCRIPT" | wc -l) #firstline is the header
    if [ "$t" != "2" ]; then
        #ensure that the script is available on or processing systems
        echo "userscript $USERSCRIPT must be on NFS" >&2
        exit 1
    fi
fi

case $ANONTYPE in
    (raw|flow|host|full) : ;;
    (*)
	echo "ANONTYPE should be either host, full, flow or raw" >&2
	usage >&2
    	exit 1
	;;
esac

if [ "$max_jobs" == "1" ]; then
    jobspercore="" #one job means only one process!
fi

USERSCRIPT_base=${USERSCRIPT##*/}

indir=${LANDER_USERDIR}/incoming_${ANONTYPE}	#~/incoming_(host|full|flow|raw)
INDIR=${INDIR:-$indir}				#use the one specified with the option or $indir if none
LOGFILE=${LANDER_USERDIR}/${USERSCRIPT_base}-d.log #write to this log in user's home

exec >> $LOGFILE 2>&1

lander_set_DATE

if [ -n "$verbose" ]; then
    [ -d "$indir" ] || echo "ERROR: incoming directory $indir does not exist (have you forgotten to run user_setup.sh?)">&2
    for datasource in $LANDER_SOURCES; do
	[ -d $indir/$datasource ] || { 
	    echo "ERROR: incoming directory for $datasource does not exist"
	    echo "       have you forgotten to run user_setup.sh?"
	} >&2
	[ -e $indir/$datasource/$LANDER_DELIVERYENABLEDFILE ] || {
	    echo "WARNING: delivery to $indir/$datasource is not enabled"
	    echo "         use user_lander.sh to enable it"
	} >&2
    done
fi

lander_set_JOBS_RUNNING $LOGNAME "$LANDER_USER_WORKER_SCRIPT" "$USERSCRIPT"
CURRENTRUNNING=$JOBS_RUNNING

# Piping to 'head' here cuts down on find's exec time
#+for very large directories by breaking the pipe early.
NUMFILES=$(find $INDIR/ -regextype posix-extended -regex ".*\/$file_regex" \
    | head -$USERJOB_MINFILES | wc -l)

echo "$DATE files:$NUMFILES workersbefore:$CURRENTRUNNING workersafter:$JOBS_RUNNING/max:$max_jobs"

[ "$NUMFILES" == "0" ] && exit 0

if [ $NUMFILES -lt $LANDER_USERJOB_MINFILES ] ; then
    # yet, if there are too stale files, we process them anyway
    topfile=$(find $INDIR/ -regextype posix-extended -type f \
	        -regex ".*\/$file_regex" -print | sort | head -1)
    [ -z "$topfile" ] && exit 0
    modtime=$(stat -c%Y $topfile) || exit 0 #must have been picked up
    now=$(date +%s)
    (( now - modtime < LANDER_USERJOB_MINFILEAGE )) && exit 0 #too young
fi

if [ $CURRENTRUNNING -lt $max_jobs ]; then
    #USER_WORKER_SCRIPT will loop, executing USERSCRIPT on all files
    #+it will be responsibility of the user/USERSCRIPT to clean up.

    inprocessdir=${INPROCESSDIR:-${LANDER_USERDIR}/inprocess_${ANONTYPE}}
    [ -n "$SKIPOOO" ] && args="-s"
    
    [ -n "$jobspercore" ] && args="$args -j $jobspercore"

    [ -n "$UNORDERED" ] && args="$args -u"

    [ -n "$FILE_REGEX" ] && args="$args -r $FILE_REGEX"

    args="$args -w $max_wait $INDIR $inprocessdir $USERSCRIPT"

    lander_create_job "$LANDER_USER_WORKER_SCRIPT" "$args" $LANDER_USERJOB_NICE

    lander_set_JOBS_RUNNING $LOGNAME "$LANDER_USER_WORKER_SCRIPT" "$USERSCRIPT"

    echo "$DATE files:$NUMFILES workersbefore:$CURRENTRUNNING workersafter:$JOBS_RUNNING/max:$max_jobs"
fi

